/* eslint-disable capitalized-comments */
/* eslint-env mocha */
'use strict';

const assert = require('assert');
const sinon = require('sinon'); // eslint-disable-line node/no-unpublished-require
const fs = require('fs');

const Dingus = require('../../lib/dingus');
const { DingusError } = require('../../lib/errors');
const Enum = require('../../lib/enum');

const noExpectedException = 'did not get expected exception';

describe('Dingus', function () {
  let dingus;
  beforeEach(function () {
    dingus = new Dingus();
  });
  afterEach(function () {
    sinon.restore();
  });

  describe('constructor', function () {
    it('covers', function () {
      const d = new Dingus({}, {});
      assert(d);
      assert('log' in d.logger);
    });
  }); // constructor

  describe('_normalizePath', function () {
    it('returns normal path', function () {
      const p = '/a/b/c';
      const r = dingus._normalizePath(p);
      assert.strictEqual(r, p);
    });
    it('returns normal path', function () {
      const p = '////a///b/./bar/..///c';
      const expected = '/a/b/c'
      const r = dingus._normalizePath(p);
      assert.strictEqual(r, expected);
    });
  }); // _normalizePath

  describe('_splitUrl', function () {
    const nullObject = Object.create(null);

    it('splits a simple path', function () {
      const p = '/a/b/c';
      const expected = {
        pathPart: p,
        queryParams: nullObject,
      };
      const r = dingus._splitUrl(p);
      assert.deepStrictEqual(r, expected);
    });
    it('splits a path with trailing slash preserved', function () {
      const p = '/a/b/c/';
      const expected = {
        pathPart: p,
        queryParams: nullObject,
      };
      const r = dingus._splitUrl(p);
      assert.deepStrictEqual(r, expected);
    });
    it('splits a path with trailing slash ignored', function () {
      const p = '/a/b/c/';
      const expected = {
        pathPart: p,
        queryParams: nullObject,
      };
      dingus.ignoreTrailingSlash = true;
      const r = dingus._splitUrl(p);
      assert.deepStrictEqual(r, expected);
    });
    it('splits a path with empty query string', function () {
      const p = '/a/b/c?';
      const expected = {
        pathPart: '/a/b/c',
        queryParams: nullObject,
      };
      const r = dingus._splitUrl(p);
      assert.deepStrictEqual(r, expected);
    });
    it('splits a path with query string', function () {
      const p = '/a/b/c?x=1&y=2&z';
      const expected = {
        pathPart: '/a/b/c',
        queryParams: Object.assign(Object.create(null), {
          x: '1',
          y: '2',
          z: '', // Subjective Editorial: disagree with the default querystring parser behavior here: null would be better than empty string, esp as result is null-prototyped object.
        }),
      };
      const r = dingus._splitUrl(p);
      assert.deepStrictEqual(r, expected);
    });
  }); // _splitUrl

  describe('tagContext', function () {
    let req, res, ctx;
    beforeEach(function () {
      req = {
        getHeader: sinon.stub(),
        setHeader: sinon.stub(),
      };
      res = {
        getHeader: sinon.stub(),
        setHeader: sinon.stub(),
      };
      ctx = {};
    });
    it ('sets id in context', function () {
      const result = Dingus.tagContext(req, res, ctx);
      assert.strictEqual(ctx.requestId, result);
      assert(res.setHeader.called);
    });
    it ('sets provided header', function () {
      req.getHeader.onCall(0).returns('abc'); // X-Request-ID
      const result = Dingus.tagContext(req, res, ctx);
      assert.strictEqual(ctx.requestId, result);
      assert.strictEqual(res.setHeader.getCall(0).args[0], 'Request-ID');
      assert.strictEqual(res.setHeader.getCall(1).args[0], 'X-Request-ID');
      assert.strictEqual(res.setHeader.getCall(1).args[1], 'abc');
      assert.strictEqual(res.setHeader.callCount, 2);
    });
  }); // tagContext

  describe('clientAddressContext', function () {
    let req, res, ctx;
    let _tp;
    before(function () {
      _tp = dingus.trustProxy;
    });
    after(function () {
      dingus.trustProxy = _tp;
    });
    beforeEach(function () {
      req = {
        getHeader: sinon.stub(),
        setHeader: sinon.stub(),
        connection: {},
      };
      res = {
        getHeader: sinon.stub(),
        setHeader: sinon.stub(),
      };
      ctx = {};
    });
    it ('covers untrusted proxy', function () {
      dingus.trustProxy = false;
      const expected = {
        clientAddress: '',
        clientProtocol: 'http',
      }
      dingus.clientAddressContext(req, res, ctx);
      assert.deepStrictEqual(ctx, expected);
      assert(!req.getHeader.called);
    });
    it ('covers missing', function () {
      dingus.trustProxy = true;
      const expected = {
        clientAddress: '::1',
        clientProtocol: 'https',
      }
      req.connection.remoteAddress = '::1';
      req.connection.encrypted = true;
      dingus.clientAddressContext(req, res, ctx);
      assert(req.getHeader.called);
      assert.deepStrictEqual(ctx, expected);
    });
  }); // clientAddressContext

  describe('getRequestContentType', function () {
    let req;
    beforeEach(function () {
      req = {
        getHeader: sinon.stub(),
        setHeader: sinon.stub(),
      };
    });
    it('handles missing header', function () {
      const result = Dingus.getRequestContentType(req);
      assert.strictEqual(result, '');
    });
    it('parses simple type', function () {
      req.getHeader.onCall(0).returns(Enum.ContentType.ApplicationJson);
      const result = Dingus.getRequestContentType(req);
      assert.strictEqual(result, Enum.ContentType.ApplicationJson);
    });
    it('parses complex type', function () {
      req.getHeader.onCall(0).returns('application/json ; charset=UTF-8');
      const result = Dingus.getRequestContentType(req);
      assert.strictEqual(result, Enum.ContentType.ApplicationJson);
    });
  }); // getRequestContentType

  describe('setResponseContentType', function () {
    let req, responseTypes;
    beforeEach(function () {
      responseTypes = [];
      req = {
        setHeader: sinon.stub(),
        getHeader: sinon.stub(),
      };
    });
    it('handles missing header', function () {
      const result = Dingus.getResponseContentType(responseTypes, req);
      assert.strictEqual(result, undefined);
    });
    it('behaves as expected', function () {
      responseTypes.push(Enum.ContentType.ApplicationJson);
      req.getHeader.onCall(0).returns('text, image/png;q=0.5, application/*;q=0.2, audio;q=0.1');
      const result = Dingus.getResponseContentType(responseTypes, req);
      assert.strictEqual(result, Enum.ContentType.ApplicationJson);
    });
  }); // setResponseContentType

  describe('on', function () {
    let stubOn;
    beforeEach(function () {
      stubOn = sinon.stub(dingus.router, 'on');
    });
    it('covers', function () {
      dingus.on('GET', '/', () => {});
      assert(stubOn.called);
    });
  }); // on

  describe('setEndBodyHandler', function () {
    let req, res, ctx, handler, origEnd, origWrite;
    beforeEach(function () {
      origEnd = sinon.stub();
      origWrite = sinon.stub();
      req = {};
      res = {
        write: origWrite,
        end: origEnd,
      };
      ctx = {};
      handler = sinon.stub();
    });
    it('collects body and handles', function () {
      Dingus.setEndBodyHandler(req, res, ctx, handler);
      res.write(Buffer.from('foo'));
      res.write('baz');
      res.write();
      res.end('quux');
      assert(origWrite.called);
      assert(origEnd.called);
      assert.deepStrictEqual(ctx.responseBody, Buffer.from('foobazquux'));
      assert(handler.called);
    });
  }); // setEndBodyHandler

  describe('setHeadHandler', function () {
    let req, res, ctx, origEnd, origWrite;
    beforeEach(function () {
      origEnd = sinon.stub();
      origWrite = sinon.stub();
      req = {
        method: 'HEAD',
      };
      res = {
        end: origEnd,
        write: origWrite,
        setHeader: sinon.stub(),
      };
      ctx = {};
    });
    it('collects body without writing', function () {
      Dingus.setHeadHandler(req, res, ctx);
      res.write(Buffer.from('foo'));
      res.write('baz');
      res.write();
      res.end('quux');
      assert(!origWrite.called);
      assert(origEnd.called);
      assert.deepStrictEqual(ctx.responseBody, Buffer.from('foobazquux'));
    });
    it('ignores non-head method', function () {
      req.method = 'GET';
      Dingus.setHeadHandler(req, res, ctx);
      res.write(Buffer.from('foo'));
      res.end('bar');
      assert(origWrite.called);
      assert(origEnd.called);
    });
  }); // setHeadHandler

  describe('addEncodingHeader', function () {
    let res, encoding;
    beforeEach(function () {
      res = {
        _headers: {},
        // eslint-disable-next-line security/detect-object-injection
        getHeader: (h) => res._headers[h],
        // eslint-disable-next-line security/detect-object-injection
        setHeader: (h, v) => res._headers[h] = v,
      };
    });
    it('adds', function () {
      encoding = 'gzip';
      Dingus.addEncodingHeader(res, encoding);
      assert.strictEqual(res._headers[Enum.Header.ContentEncoding], 'gzip');
    });
    it('extends', function () {
      encoding = 'utf8';
      Dingus.addEncodingHeader(res, encoding);
      assert.strictEqual(res._headers[Enum.Header.ContentEncoding], 'utf8');
      encoding = 'gzip';
      Dingus.addEncodingHeader(res, encoding);
      assert.strictEqual(res._headers[Enum.Header.ContentEncoding], 'gzip, utf8');
    });
  }); // addEncodingHeader

  describe('dispatch', function () {
    let pathsByLengthOrig;
    let req, res, ctx;
    let stubHandler;

    beforeEach(function () {
      req = {
        url: '/',
        method: 'GET',
        setHeader: sinon.stub(),
        getHeader: sinon.stub(),
      };
      res = {
        statusCode: 200,
        end: sinon.stub(),
        setHeader: sinon.stub(),
        hasHeader: sinon.stub(),
        getHeader: sinon.stub(),
        getHeaders: sinon.stub(),
      };
      ctx = {};
      pathsByLengthOrig = dingus.pathsByLength;
      sinon.spy(dingus, 'handlerMethodNotAllowed');
      sinon.spy(dingus, 'handlerNotFound');
      sinon.spy(dingus, 'handlerBadRequest');
      sinon.spy(dingus, 'handlerInternalServerError');
      stubHandler = sinon.stub();
    });
    afterEach(function () {
      dingus.pathsByLength = pathsByLengthOrig;
    });

    it('calls handler', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      dingus.on(method, urlPath, stubHandler);
      req.url = '/abc';
      req.method = method;

      await dingus.dispatch(req, res, ctx);
      assert(stubHandler.called);
      assert(!dingus.handlerMethodNotAllowed.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('calls handler without context', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      dingus.on(method, urlPath, stubHandler);
      req.url = '/abc';
      req.method = method;

      await dingus.dispatch(req, res);
      assert(stubHandler.called);
      assert(!dingus.handlerMethodNotAllowed.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('calls fallback handler', async function () {
      const urlPath = '/abc/:id';
      const method = 'GET';
      dingus.on('*', urlPath, stubHandler);
      req.url = '/abc/def';
      req.method = method;

      await dingus.dispatch(req, res, ctx);
      assert(stubHandler.called);
      assert(!dingus.handlerMethodNotAllowed.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('handles error in handler', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      dingus.on(method, urlPath, stubHandler);
      req.url = '/abc';
      req.method = method;
      stubHandler.rejects(new Error('blah'));

      await dingus.dispatch(req, res, ctx);
      assert(stubHandler.called);
      assert(!dingus.handlerMethodNotAllowed.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('calls unsupported method', async function () {
      const urlPath = '/:id';
      const method = 'POST';
      dingus.on('GET', urlPath, stubHandler);
      req.url = '/abc';
      req.method = method;

      await dingus.dispatch(req, res, ctx);
      assert(!stubHandler.called);
      assert(dingus.handlerMethodNotAllowed.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('does not lookup nonexistent path', async function () {
      req.url = '/foo/bar';
      req.method = 'GET';

      await dingus.dispatch(req, res, ctx);
      assert(!stubHandler.called);
      assert(!dingus.handlerMethodNotAllowed.called);
      assert(dingus.handlerNotFound.called);
    });
    it('covers unhandled dingus exception', async function () {
      const expectedException = new DingusError('blah');
      sinon.stub(dingus.router, 'lookup').throws(expectedException);

      await dingus.dispatch(req, res, ctx);
      assert(!stubHandler.called);
      assert(dingus.handlerInternalServerError.called);
    });
    it('covers other exception', async function () {
      const expectedException = new Error('blah');
      sinon.stub(dingus.router, 'lookup').throws(expectedException);

      await dingus.dispatch(req, res, ctx);
      assert(!stubHandler.called);
      assert(dingus.handlerInternalServerError.called);
    });
    it('covers bad uri', async function () {
      req.url = '/%f';

      await dingus.dispatch(req, res, ctx);
      assert(dingus.handlerBadRequest.called);
    });
    it('calls handler with additional arguments', async function () {
      dingus.on('GET', '/', stubHandler, 'foo', 'bar');
      await dingus.dispatch(req, res, ctx);
      assert(stubHandler.called);
      assert.strictEqual(stubHandler.args[0][3], 'foo');
      assert.strictEqual(stubHandler.args[0][4], 'bar');
    });
  }); // dispatch

  describe('parseBody', function () {
    let ctx;
    beforeEach(function () {
      ctx = {};
    });
    it('does not parse unknown type', function () {
      try {
        dingus.parseBody('unknown/type', ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.statusCode, 415);
      }
    });
    it('parses json', function () {
      const src = { foo: 'bar' };
      ctx.rawBody = JSON.stringify(src);
      dingus.parseBody(Enum.ContentType.ApplicationJson, ctx);
      assert.deepStrictEqual(ctx.parsedBody, src);
    });
    it('handles unparsable json', function () {
      ctx.rawBody = 'not json';
      try {
        dingus.parseBody(Enum.ContentType.ApplicationJson, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.statusCode, 400);
      }
    });
    it('parses form', function () {
      const expected = Object.assign(Object.create(null), {
        foo: 'bar',
      });
      ctx.rawBody = 'foo=bar';
      dingus.parseBody('application/x-www-form-urlencoded', ctx);
      assert.deepStrictEqual(ctx.parsedBody, expected);
    });

  }); // parseBody

  describe('bodyData', function () {
    let res, resEvents;
    beforeEach(function () {
      resEvents = {};
      res = {
        // eslint-disable-next-line security/detect-object-injection
        on: (ev, fn) => resEvents[ev] = fn,
      };
    });
    it('provides data', async function () {
      const p = dingus.bodyData(res);
      resEvents['data'](Buffer.from('foo'));
      resEvents['data'](Buffer.from('bar'));
      resEvents['end']();
      const result = await p;
      assert.strictEqual(result, 'foobar');
    });
    it('handles error', async function () {
      const p = dingus.bodyData(res);
      resEvents['error']('foo');
      try {
        await p;
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e, 'foo');
      }
    });
  }); // bodyData

  describe('ingestBody', function () {
    it('covers', async function () {
      const req = {};
      const res = {};
      const ctx = {};
      sinon.stub(dingus, 'bodyData').resolves('{"foo":"bar"}')
      sinon.stub(Dingus, 'getRequestContentType').returns(Enum.ContentType.ApplicationJson);
      await dingus.ingestBody(req, res, ctx);
      assert.deepStrictEqual(ctx.parsedBody, { foo: 'bar' });
    });
  }); // ingestBody

  describe('setResponseType', function () {
    let req, res, ctx;
    let _sa; // Preserve strictAccept
    before(function () {
      _sa = dingus.strictAccept;
    });
    after(function () {
      dingus.strictAccept = _sa;
    });
    beforeEach(function () {
      ctx = {};
      req = {};
      res = {
        setHeader: sinon.stub(),
      };
      sinon.stub(Dingus, 'getResponseContentType').returns();
    });
    it('rejects missing', function () {
      dingus.strictAccept = true;
      try {
        dingus.setResponseType(['my/type'], req, res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.statusCode, 406, 'did not get expected status code');
      }
    });
    it('accepts missing', function () {
      dingus.strictAccept = false;
      dingus.setResponseType(['my/type'], req, res, ctx);
      assert.strictEqual(ctx.responseType, 'my/type');
    });

  }); // setResponseType

  describe('_readFileInfo', function () {
    let stat, data, statRes, dataRes, filename;
    beforeEach(function () {
      sinon.stub(fs.promises, 'stat');
      sinon.stub(fs.promises, 'readFile');
      statRes = {
        mtimeMs:1612553697186,
      };
      dataRes = 'data';
      filename = 'dummy.txt';
    });
    it('succeeds', async function () {
      fs.promises.stat.resolves(statRes);
      fs.promises.readFile.resolves('data');
      [stat, data] = await dingus._readFileInfo(filename);
      assert.deepStrictEqual(stat, statRes);
      assert.deepStrictEqual(data, dataRes);
    });
    it('returns null for non-existant file', async function () {
      const noEnt = {
        code: 'ENOENT',
      };
      fs.promises.stat.rejects(noEnt);
      fs.promises.readFile.rejects(noEnt);
      [stat, data] = await dingus._readFileInfo(filename);
      assert.strictEqual(stat, null);
      assert.strictEqual(data, null);
    });
    it('throws unexpected error', async function () {
      const expectedException = new Error('blah');
      fs.promises.stat.rejects(expectedException);
      await assert.rejects(async () => {
        await dingus._readFileInfo(filename);
      }, expectedException);
    });
  }); // _readFileInfo

  describe('_serveFileMetaHeaders', function () {
    let res, directory, fileName;
    beforeEach(function () {
      sinon.stub(dingus, '_readFileInfo');
      res = {
        setHeader: sinon.stub(),
      };
      directory = '/path';
      fileName = 'filename';
    });
    it('covers no meta file', async function() {
      dingus._readFileInfo.resolves([null, null]);
      await dingus._serveFileMetaHeaders(res, directory, fileName);
      assert(!res.setHeader.called);
    });
    it('adds extra headers', async function () {
      dingus._readFileInfo.resolves([{}, Buffer.from(`Link: <https://example.com/>; rel="relation"
X-Folded-Header: data
  data under
  the fold
Content-Type: image/sgi
`)]);
      await dingus._serveFileMetaHeaders(res, directory, fileName);
      assert(res.setHeader.called);
    });
  }); // _serveFileMetaHeaders

  describe('serveFile', function () {
    const path = require('path');
    let ctx, req, res, directory, fileName, filestats;
    beforeEach(function () {
      directory = path.join(__dirname, '..', 'test-data');
      fileName = 'example.html';
      ctx = {};
      req = {
        _headers: {
          [Enum.Header.Accept]: undefined,
          [Enum.Header.IfModifiedSince]: undefined,
          [Enum.Header.AcceptEncoding]: undefined,
          [Enum.Header.IfNoneMatch]: undefined,
        },
        getHeader: (header) => {
          if (header in req._headers) {
            // eslint-disable-next-line security/detect-object-injection
            return req._headers[header];
          }
          assert.fail(`unexpected getHeader ${header}`);
        },
      };
      res = {
        end: sinon.stub(),
        getHeader: sinon.stub(),
        getHeaders: sinon.stub(),
        hasHeader: sinon.stub().returns(true),
        setHeader: sinon.stub(),
      };
      filestats = {
        dev: 39,
        mode: 33188,
        nlink: 1,
        uid: 1002,
        gid: 1002,
        rdev: 0,
        blksize: 512,
        ino: 897653,
        size: 8,
        blocks: 17,
        atimeMs: 1613253436842.815,
        mtimeMs: 1603485933192.861,
        ctimeMs: 1603485933192.861,
        birthtimeMs: 0,
        atime: '2021-02-13T21:57:16.843Z',
        mtime: '2020-10-23T13:45:33.193Z',
        ctime: '2020-10-23T13:45:33.193Z',
        birthtime: '1970-01-01T00:00:00.000Z',
      };
      sinon.stub(dingus, 'handlerNotFound');
      sinon.stub(fs.promises, 'stat').resolves(filestats);
      sinon.spy(fs.promises, 'readFile');
    });
    it('serves a file', async function () {
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(fs.promises.readFile.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('covers no meta headers', async function () {
      dingus.staticMetadata = false;
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(fs.promises.readFile.called);
      assert(!dingus.handlerNotFound.called);
    });
    it('does not serve dot-file', async function () {
      fileName = '.example';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(!fs.promises.readFile.called);
      assert(dingus.handlerNotFound.called);
    });
    it('does not serve encoded navigation', async function () {
      fileName = '/example.html';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(!fs.promises.readFile.called);
      assert(dingus.handlerNotFound.called);
    });
    it('does not serve missing file', async function () {
      fileName = 'no-file.here';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(dingus.handlerNotFound.called);
    });
    it('requires directory be specified', async function () {
      await dingus.serveFile(req, res, ctx, '', fileName);
      assert(!fs.promises.readFile.called);
      assert(dingus.handlerNotFound.called);
    });
    it('covers fs error', async function () {
      const expectedException = new Error('blah');
      fs.promises.stat.restore();
      sinon.stub(fs.promises, 'stat').rejects(expectedException);
      try {
        await dingus.serveFile(req, res, ctx, directory, fileName);
        assert.fail('should have thrown');
      } catch (e) {
        assert.strictEqual(e, expectedException);
      }
    });
    it('caches by modified', async function () {
      req._headers[Enum.Header.IfModifiedSince] = 'Fri, 23 Oct 2020 23:11:16 GMT';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert.strictEqual(res.statusCode, 304);
    });
    it('does not cache old modified', async function () {
      req._headers[Enum.Header.IfModifiedSince] = 'Fri, 23 Oct 2020 01:11:16 GMT';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert.notStrictEqual(res.statusCode, 304);
      assert(!dingus.handlerNotFound.called);
    });
    it('caches ETag match', async function () {
      req._headers[Enum.Header.IfNoneMatch] = '"zPPQVfXV36sgXq4fRLdsm+7rRMb8IUfb/eJ6N6mnwWs"';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert.strictEqual(res.statusCode, 304);
    });
    it('does not cache ETag non-match', async function () {
      req._headers[Enum.Header.IfNoneMatch] = '"foo", "bar"';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert.notStrictEqual(res.statusCode, 304);
      assert(!dingus.handlerNotFound.called);
    });
    it('handles no possible encodings', async function () {
      req._headers[Enum.Header.AcceptEncoding] = '*;q=0';
      await assert.rejects(async () => {
        await dingus.serveFile(req, res, ctx, directory, fileName);
      }, {
        name: 'ResponseError',
      });
    });
    it('handles a valid encoding', async function () {
      req._headers[Enum.Header.AcceptEncoding] = 'gzip';
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(res.end.called);
    });
    it('handles a valid encoding among others', async function () {
      req._headers[Enum.Header.AcceptEncoding] = 'flarp, br, gzip';
      fs.promises.stat.restore();
      sinon.stub(fs.promises, 'stat')
        .onCall(0).resolves(filestats) // identity file
        .onCall(1).resolves(null) // br encoding
        .onCall(2).resolves(filestats); // gzip encoding
      await dingus.serveFile(req, res, ctx, directory, fileName);
      assert(res.end.called);
    });
  }); // serveFile

  describe('renderError', function () {
    let err;
    beforeEach(function () {
      err = {
        statusCode: '200',
        errorMessage: 'OK',
        details: 'hunkydorey',
      };
    });
    it('renders unknown type', function () {
      const contentType = 'unknown/type';
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, 'OK\r\nhunkydorey');
    });
    it('renders text', function () {
      const contentType = 'text/plain';
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, 'OK\r\nhunkydorey');
    });
    it('renders json', function () {
      const contentType = Enum.ContentType.ApplicationJson;
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, JSON.stringify(err));
    });
    it('renders html without details', function () {
      err = {
        statusCode: '201',
        errorMessage: 'Created',
      };
      const contentType = 'text/html';
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, `<!DOCTYPE html>
<html lang="en">
  <head>
    <title>${err.statusCode} ${err.errorMessage}</title>
  </head>
  <body>
    <h1>${err.errorMessage}</h1>
  </body>
</html>`);
    });
    it('renders html', function () {
      const contentType = 'text/html';
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, `<!DOCTYPE html>
<html lang="en">
  <head>
    <title>${err.statusCode} ${err.errorMessage}</title>
  </head>
  <body>
    <h1>${err.errorMessage}</h1>
    <p>${err.details}</p>
  </body>
</html>`);
    });
    it('renders html, multiple details', function () {
      const contentType = 'text/html';
      err.details = ['one detail', 'two detail'];
      const result = dingus.renderError(contentType, err);
      assert.deepStrictEqual(result, `<!DOCTYPE html>
<html lang="en">
  <head>
    <title>${err.statusCode} ${err.errorMessage}</title>
  </head>
  <body>
    <h1>${err.errorMessage}</h1>
    <p>one detail</p>
    <p>two detail</p>
  </body>
</html>`);
    });
  }); // renderError

  describe('sendErrorResponse', function () {
    let ctx, req, res;
    beforeEach(function () {
      ctx = {};
      req = {};
      res = {
        end: sinon.stub(),
        getHeader: sinon.stub(),
        getHeaders: sinon.stub(),
        hasHeader: sinon.stub().returns(true),
        setHeader: sinon.stub(),
      };
      sinon.stub(dingus, 'renderError');
    });
    it('covers', function () {
      const err = {
        statusCode: 444,
      };
      dingus.sendErrorResponse(err, req, res, ctx);
      assert(res.end.called);
    });
  }); // sendErrorResponse

  describe('proxyPrefix', function () {
    let req, res, ctx, stubHandler, pfxDingus;
    const pfx = '/pfx';

    beforeEach(function () {
      pfxDingus = new Dingus(console, { proxyPrefix: pfx });
      req = {
        setHeader: sinon.stub(),
        getHeader: sinon.stub(),
      };
      res = {
        statusCode: 200,
        end: sinon.stub(),
        setHeader: sinon.stub(),
        getHeader: sinon.stub(),
      };
      ctx = {};
      sinon.stub(pfxDingus, 'handlerMethodNotAllowed');
      sinon.stub(pfxDingus, 'handlerNotFound');
      stubHandler = sinon.stub();
    });
    afterEach(function () {
      sinon.restore();
    });

    it('handles prefixed route', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      pfxDingus.on(method, urlPath, stubHandler);
      req.url = pfx + '/abc';
      req.method = method;

      await pfxDingus.dispatch(req, res, ctx);
      assert(stubHandler.called);
      assert(!pfxDingus.handlerMethodNotAllowed.called);
      assert(!pfxDingus.handlerNotFound.called);
    });
    it('does not handle prefixed route', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      pfxDingus.on(method, urlPath, stubHandler);
      req.url = '/wrongpfx/abc';
      req.method = method;

      await pfxDingus.dispatch(req, res, ctx);
      assert(!stubHandler.called);
      assert(!pfxDingus.handlerMethodNotAllowed.called);
      assert(pfxDingus.handlerNotFound.called);
    });
  }); // proxyPrefix

  describe('handlerRedirect', function () {
    let req, res, ctx;
    beforeEach(function () {
      req = {
        getHeader: sinon.stub(),
      };
      res = {
        setHeader: sinon.stub(),
        end: sinon.stub(),
      };
      ctx = {};
    });
    it('covers', async function () {
      await dingus.handlerRedirect(req, res, ctx);
      assert(res.setHeader.called);
      assert(res.end.called);
    });
    it('covers non-defaults', async function () {
      await dingus.handlerRedirect(req, res, ctx, 308);
      assert(res.setHeader.called);
      assert(res.end.called);
    });
  }); // handlerRedirect

  describe('handlerGetStaticFile', function () {
    let req, res, ctx;
    beforeEach(function () {
      req = {
        getHeader: sinon.stub(),
      };
      res = {
        setHeader: sinon.stub(),
      };
      ctx = {
        params: {
          file: '',
        },
      };
      sinon.stub(dingus, 'serveFile');
    });
    it('covers', async function () {
      await dingus.handlerGetStaticFile(req, res, ctx);
      assert(dingus.serveFile.called);
    });
    it('covers specified file', async function () {
      await dingus.handlerGetStaticFile(req, res, ctx, 'file.txt');
      assert(dingus.serveFile.called);
    });
  }); // handlerGetStaticFile
});